import os
from io import BytesIO

import docx
from bureaucracy.tests.test_fields import DocxTestsBase, resources_dir
from PyPDF2.pdf import PdfFileReader


class RenderTests(DocxTestsBase):
    def setUp(self):
        self.out_path = os.path.join(resources_dir, 'tmp.docx')
        self.out_path_pdf = os.path.join(resources_dir, 'tmp.pdf')


    def tearDown(self):
        if os.path.exists(self.out_path):
            os.remove(self.out_path)

        if os.path.exists(self.out_path_pdf):
            os.remove(self.out_path_pdf)

    def test_save_as_docx(self):
        doc = self._get_docx('complex_fields')
        doc.render_and_save(self.out_path,
                            context={'complex': 'BEEES. AAAAH. BEEEEES',
                                     'complex2': 'ಠ_ಠ unifying matrix conventions is the way of the future, Max.'})

        self.assertTrue(os.path.exists(self.out_path))

        # can the pytohn-docx library parse our result without throwing a hissy fit?
        docx.Document(self.out_path)

    def test_save_as_pdf(self):
        doc = self._get_docx('complex_fields')
        doc.render_and_save(self.out_path_pdf,
                            context={'complex': 'BEEES. AAAAH. BEEEEES',
                                     'complex2': 'ಠ_ಠ unifying matrix conventions is the way of the future, Max.'},
                            format='pdf')

        self.assertTrue(os.path.exists(self.out_path_pdf))

        # does pypdf like our output?
        pdf = open(self.out_path_pdf, "rb")
        PdfFileReader(pdf)
        pdf.close()

    def test_render_to_pdf_bytes(self):
        doc = self._get_docx('complex_fields')
        data = doc.render(context={'complex': 'BEEES. AAAAH. BEEEEES',
                                   'complex2': 'ಠ_ಠ unifying matrix conventions is the way of the future, Max.'},
                          format='pdf')

        # does pypdf like our output?
        PdfFileReader(BytesIO(data))

    def test_render_to_docx_bytes(self):
        doc = self._get_docx('complex_fields')
        data = doc.render(context={'complex': 'BEEES. AAAAH. BEEEEES',
                                            'complex2': 'ಠ_ಠ unifying matrix conventions is the way of the future, Max.'})

        # can the python-docx library parse our result without throwing a hissy fit?
        docx.Document(BytesIO(data))



import os
import unittest

from bureaucracy import DocxTemplate

resources_dir = os.path.join(os.path.realpath(os.path.dirname(__file__)), 'resources')


class DocxTestsBase(unittest.TestCase):
    def _get_docx(self, name):
        return DocxTemplate('{}/{}.docx'.format(resources_dir, name))


class GetFieldNamesTests(DocxTestsBase):
    def test_get_only_fldChar(self):
        doc = self._get_docx('complex_fields')
        self.assertEqual({'complex', 'complex2'}, doc.get_field_names())

    def test_get_only_fldSimple(self):
        doc = self._get_docx('simple_fields')
        self.assertEqual({'foo', 'bar', 'baz'}, doc.get_field_names())

    def test_get_all_fields(self):
        doc = self._get_docx('simple_and_complex_fields')
        self.assertEqual({'foo', 'bar', 'baz', 'complex', 'complex2', 'simple'}, doc.get_field_names())


class FieldTests(DocxTestsBase):
    def test_replace_only_fldSimple(self):
        doc = self._get_docx('simple_fields')
        doc.replace_fields({'foo': 'frobnicate',
                            'bar': 'determiorate',
                            'baz': 'some kind of string'})

        self.assertEqual(len(doc.get_field_names()), 0)

        self.assertTrue(doc._element.xpath(".//text()='frobnicate'"))
        self.assertTrue(doc._element.xpath(".//text()='determiorate'"))
        self.assertTrue(doc._element.xpath(".//text()='some kind of string'"))

    def test_replace_only_fldChar(self):
        doc = self._get_docx('complex_fields')
        doc.replace_fields({'complex': 'BEEES. AAAAH. BEEEEES',
                            'complex2': 'ಠ_ಠ unifying matrix conventions is the way of the future, Fred.'})

        self.assertEqual(len(doc.get_field_names()), 0)
        self.assertTrue(doc._element.xpath(".//text()='BEEES. AAAAH. BEEEEES'"))
        self.assertTrue(doc._element.xpath(".//text()='ಠ_ಠ unifying matrix conventions is the way of the future, Fred.'"))

    def test_replace_all_fields(self):
        doc = self._get_docx('simple_and_complex_fields')
        doc.replace_fields({'foo': 'frobnicate',
                            'bar': 'determiorate',
                            'baz': 'some kind of string',
                            'complex': 'BEEES. AAAAH. BEEEEES',
                            'complex2': 'ಠ_ಠ unifying matrix conventions is the way of the future, Fred.',
                            'simple': 5.2})

        self.assertEqual(len(doc.get_field_names()), 0)

        self.assertTrue(doc._element.xpath(".//text()='frobnicate'"))
        self.assertTrue(doc._element.xpath(".//text()='determiorate'"))
        self.assertTrue(doc._element.xpath(".//text()='some kind of string'"))
        self.assertTrue(doc._element.xpath(".//text()='BEEES. AAAAH. BEEEEES'"))
        self.assertTrue(doc._element.xpath(".//text()='ಠ_ಠ unifying matrix conventions is the way of the future, Fred.'"))
        self.assertTrue(doc._element.xpath(".//text()='5.2'"))



import itertools
import logging
import os
import re
import subprocess
import tempfile
from copy import deepcopy
from io import BytesIO

from bureaucracy.replacements import (HTMLReplacement, ImageReplacement,
                                      Replacement, TableReplacement,
                                      TextReplacement)
from bureaucracy.utils import namespaced
from docx.document import Document
from docx.opc.constants import CONTENT_TYPE
from docx.package import Package
from docx.text.paragraph import Paragraph
from docx.text.run import Run
from lxml.etree import tostring

r = re.compile(r' MERGEFIELD +"?([^ ]+?)"? +(|\\\* MERGEFORMAT )', re.I)  # fixme. it might be not a simple as that
# the ooxml standard says that the instrText can be divided over several runs. i have not been able to make microsoft
# word generate something like this though...

logger = logging.getLogger('bureaucracy')


class DocxTemplate(Document):
    def __init__(self, docx, strict=False):
        """
        Initialize the DocxTemplate with the given docx file or stream.
        :param docx: A string or file like object (django.core.File objects work) representing a docx file.
        :param strict: will make the template render throw an error when fields and context do not match if True
        """
        self.strict = strict
        document_part = Package.open(docx).main_document_part
        if document_part.content_type != CONTENT_TYPE.WML_DOCUMENT_MAIN:
            tmpl = "file '%s' is not a Word file, content type is '%s'"
            raise ValueError(tmpl % (docx, document_part.content_type))
        super().__init__(document_part._element, document_part)

    def get_field_names(self):
        """
        Get the name of the mailmerge fields included in the document
        :return: a set of the fields' names
        """
        return {rv[0] for rv in self.iter_fields()}

    def iter_fields(self):
        """
        Gernerator for fldSimple and instrText fields and their fieldnames
        :return: a generator yielding tuples (field name, field)-tuples.
        """
        # there are two ways a mergefield can be represented: the simple way with fldSimple and the more
        # complex way with instrText and fldChar.

        simple_fields = ((field.attrib[namespaced('instr')], field) for field in self._element.xpath('.//w:fldSimple'))
        complex_fields = ((field.text, field) for field in self._element.xpath('.//w:instrText'))

        for instr, field in itertools.chain(simple_fields, complex_fields):

            m = r.match(instr)
            if not m and self.strict:
                raise ValueError("Could not determine name of merge field with instr text '{}'".format(instr))
            elif not m:
                logger.warning("Could not determine name of merge field with instr text '{}'. Skipping".format(instr))
                continue

            yield m.group(1), field

    def replace_simple_field(self, field, replacement):

        # a fldSimple tag is easily replaced, we just create a new run in the same paragraph and replace that one
        # with the fldSimple node.

        parent_node = field.getparent()

        # the standard says that this is the case most of the time so we only deal with this case for now:
        assert parent_node.tag == namespaced('p')

        current_paragraph = Paragraph(parent_node, self._body)
        replacement_run = Run(current_paragraph._p._add_r(), current_paragraph)
        parent_node.replace(field, replacement_run._element)
        replacement.fill(replacement_run)

    def replace_complex_field(self, field, replacement):
        # fldChar is more complex. it's not a tag, but rather a series of fldChar and instrText tags inside separate
        # runs. The tags that concern us are these:
        #
        #  1. <w:fldChar w:fldCharType="begin"/>: The beginning of the field
        #  2. <w:instrText xml:space="preserve"> MERGEFIELD test \* MERGEFORMAT </w:instrText> contains the field's name
        #  3. <w:fldChar w:fldCharType="end"/> Marks the end of the field
        #

        # get the run this field is in
        instr_run_node = field.getparent()
        assert instr_run_node.tag == namespaced('r')

        # we now look for the run containing of the opening fldChar for this instrText, which is the first one
        # with an opening fldChar we encounter before the run with instrText
        opening_run_node = instr_run_node
        while not opening_run_node.xpath('w:fldChar[@w:fldCharType="begin"]'):
            opening_run_node = opening_run_node.getprevious()
            if opening_run_node is None:
                raise ValueError(
                    "Could not find beginning of field with instr node '{}'?! Is the document malformed?".format(field))

        # idem for the run containing the closing fldChar, but of course now looking ahead
        closing_run_node = instr_run_node
        while not closing_run_node.xpath('w:fldChar[@w:fldCharType="end"]'):
            closing_run_node = closing_run_node.getnext()
            if closing_run_node is None:
                raise ValueError(
                    "Could not find end of field with instr node '{}'?! Is the document malformed?".format(field))

        # now replace all runs between the opening and closing runs
        current_paragraph = Paragraph(instr_run_node.getparent(), self._body)

        begin_index = current_paragraph._element.index(opening_run_node)
        end_index = current_paragraph._element.index(closing_run_node)

        run = Run(current_paragraph._p._add_r(), current_paragraph)
        current_paragraph._element[begin_index:end_index] = [run._element]

        replacement.fill(run)

    def replace_fields(self, context):
        unused_fields = set()
        unused_values = set(context.keys())

        for field_name, field in self.iter_fields():

            if field_name in context:
                unused_values.discard(field_name)
                replacement = context[field_name]
                if not isinstance(replacement, Replacement):
                    replacement = TextReplacement(replacement)
            else:
                if self.strict:
                    raise ValueError('Could not find field name {} in context'.format(field_name))
                else:
                    replacement = TextReplacement('')
                    unused_fields.add(field_name)

            if field.tag == namespaced('fldSimple'):
                self.replace_simple_field(field, replacement)
            elif field.tag == namespaced('instrText'):
                self.replace_complex_field(field, replacement)

        if unused_fields:
            logger.warn("Fields %s were present in the document, but not in the context. They were removed",
                        unused_fields)
        if unused_values:
            logger.warn(
                "Values %s were present in the context, but no corresponding fields were found in the document.",
                unused_values)

    def render(self, context, format='docx'):
        doc = deepcopy(self)  # take a copy so we can keep using this instance to generate from other contexts
        doc.replace_fields(context)

        if format == 'docx':
            handle = BytesIO()
            doc.save(handle)
            handle.seek(0)
            return handle.read()
        elif format == 'pdf':
            return doc.to_pdf_bytes()
        else:
            raise Exception('Unsupported format.')

    def render_and_save(self, path, context, format='docx'):
        doc = deepcopy(self)  # take a copy so we can keep using this instance to generate from other contexts
        doc.replace_fields(context)

        if format == 'docx':
            doc.save(path)
        elif format == 'pdf':
            doc.to_pdf(path)

    # what follows is a hack.
    # To be able to generate a pdf document/bytes from a merged document, we call libreoffice's headless
    # command line utility to to the generation. The problem is is that that tool only works with files,
    # not with bytestreams or strings, so we save this docx in a tmp dir, throw that into soffice, which
    # generates a pdf file in the same dir with the same name. Then we return bytes dumped into that file or move
    # the file to the desired path, depending on whether we're saving or rendering directly to a stream.
    # All this also makes rendering to pdf very slow. when doing a lot of renders, it might be a good idea to use
    # the batch functionality in the soffice convert tool.

    def _to_pdf(self, path=None):
        _, tmp_doc_path = tempfile.mkstemp(suffix='.docx')
        soffice_outdir = tempfile.gettempdir()
        tmp_pdf_outfilename = '{}.pdf'.format(tmp_doc_path.split(os.sep)[-1].split('.')[0])
        tmp_pdf_path = os.path.join(soffice_outdir, tmp_pdf_outfilename)

        self.save(tmp_doc_path)
        try:
            subprocess.call(['soffice', '--headless',
                             '--convert-to', 'pdf',
                             tmp_doc_path, '--outdir', soffice_outdir],
                            stdout=subprocess.DEVNULL)

            if path:
                os.rename(tmp_pdf_path, path)
            else:
                with open(tmp_pdf_path, 'rb') as f:
                    return f.read()

        finally:
            if os.path.exists(tmp_doc_path):
                os.remove(tmp_doc_path)

            if os.path.exists(tmp_pdf_path):
                os.remove(tmp_pdf_path)

    def to_pdf(self, path):
        self._to_pdf(path)

    def to_pdf_bytes(self):
        return self._to_pdf()

    def __str__(self):
        return tostring(self._element, pretty_print=True).decode('utf-8')


if __name__ == '__main__':
    examples_dir = os.path.join(os.path.realpath(os.path.dirname(__file__)), '..', 'examples')
    doc = DocxTemplate(os.path.join(examples_dir, 'sample.docx'))
    context = {
        'table': TableReplacement([[str(i * j) for j in range(1, 8)] for i in range(1, 6)],
                                  headers=['one', 'two', 'three', 'four', 'five', 'six', 'seven']),
        'image': ImageReplacement(os.path.join(examples_dir, 'pigeony.jpg')),
        'html': HTMLReplacement(
            open(os.path.join(examples_dir, '..', 'examples', 'test.html')).read()),
        'text': 'some text',
        'hop': 546
    }

    doc.render_and_save(
        os.path.join(examples_dir, 'generated.docx'),
        context)



from lxml.etree import tostring


NAMESPACES = {
    'w': 'http://schemas.openxmlformats.org/wordprocessingml/2006/main',
    'mc': 'http://schemas.openxmlformats.org/markup-compatibility/2006',
    'ct': 'http://schemas.openxmlformats.org/package/2006/content-types',
}


def print_node(node):
    print(tostring(node, pretty_print=True).decode('utf-8'))


def namespaced(name, ns='w'):
    return '{{{0}}}{1}'.format(NAMESPACES[ns], name)


DOCX_MIMETYPE = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
PDF_MIMETYPE = 'application/pdf'



import os
import tempfile
from copy import copy

import pypandoc
from docx import Document
from docx.oxml import CT_Tbl
from docx.table import Table
from docx.text.paragraph import Paragraph


class Replacement(object):
    def fill(self, el):
        raise NotImplementedError


class RunReplacement(Replacement):
    pass


class ParagraphReplacement(Replacement):
    def fill_paragraph(self, par):
        raise NotImplementedError

    def fill(self, run):
        paragraph = Paragraph(run._element.getparent(), run._parent._parent)
        self.fill_paragraph(paragraph)


class ImageReplacement(RunReplacement):
    def __init__(self, filename, width=None, height=None):
        self.filename = filename
        self.width = width
        self.height = height

    def fill(self, run):
        run.add_picture(self.filename, self.width, self.height)


class HTMLReplacement(ParagraphReplacement):
    def __init__(self, html):
        # fixme: pandoc doesn't allow us to generate a docx stream in memory, so we have to use a temporary file,
        # which makes this part slow in comparison to other replacements. if anyone knows about a good html to
        # docx converter; help yourself.

        _, tmp_file_name = tempfile.mkstemp(suffix='.docx')
        pypandoc.convert(html, format='html', to='docx', outputfile=tmp_file_name)
        doc = Document(tmp_file_name)

        # find all paragraphs in the output
        self.par_nodes = doc._element.xpath('./w:body/w:p')

        # we just add all styles in the output document. This opens us up to the possibility of clashing styles,
        # but for now that doesn't seem to be an issue. only adding the styles used in the output of pandoc and their
        # ancestors in the style hierarchy would be the ideal solutionm, but this works for now.
        self.styles = doc.styles

        os.remove(tmp_file_name)

    def fill_paragraph(self, par):
        body_el = par._element.getparent()
        par_idx = body_el.index(par._element)
        body_el[par_idx:par_idx + 1] = self.par_nodes

        for style in self.styles:
            par.part.document.styles._element.append(copy(style._element))


class TableReplacement(ParagraphReplacement):
    def __init__(self, data, headers):
        if not isinstance(data, (list, tuple)) or any(not isinstance(d, (list, tuple)) for d in data):
            raise Exception("data should be a list or tuple of lists or tuples")
        if headers and not isinstance(headers, (list, tuple)):
            raise Exception("headers should be a list or tuple")

        if isinstance(data, (list, tuple)) and data:
            if any(len(d) != len(data[0]) for d in data):
                raise ValueError("Data contains rows of varying sizes")
            if headers and len(data[0]) != len(headers):
                raise ValueError("Data and header lengths do not match")

        self.headers = headers
        self.data = data

    def fill_paragraph(self, par):

        nr_rows = len(self.data) if self.data else 0
        if self.headers:
            nr_rows += 1

        nr_cols = len(self.data[0]) if self.data else len(self.headers) if self.headers else 0

        # replace the par element with a table xml element and build a docx.Table from it
        # so we can use that to fill it up
        table_el = CT_Tbl.new_tbl(nr_rows, nr_cols, par.part.document._block_width)
        par._element.getparent().replace(par._element, table_el)
        table = Table(table_el, self)

        if self.headers:
            for i, header in enumerate(self.headers):
                table.rows[0].cells[i].text = str(header)

        for row_idx, row_values in enumerate(self.data, start=1 if self.headers else 0):
            for col_idx, cell_value in enumerate(row_values):
                table.rows[row_idx].cells[col_idx].text = str(cell_value)


class TextReplacement(RunReplacement):
    def __init__(self, text):
        self.text = text

    def fill(self, run):
        run.text = str(self.text)



import os
import re
from copy import copy
from zipfile import ZipFile

from bureaucracy import HTML, Image, Table
from bureaucracy.tests.test_fields import DocxTestsBase, resources_dir
from docx.enum.style import WD_STYLE_TYPE


class ImageReplacementTests(DocxTestsBase):
    def setUp(self):
        self.generated_path = os.path.join(resources_dir, 'generated_image.docx')

    def tearDown(self):
        if os.path.exists(self.generated_path):
            os.remove(self.generated_path)

    def test_replace_image(self):
        doc = copy(self._get_docx('image'))

        doc.replace_fields({
            'image': Image(os.path.join(resources_dir, 'pigeon.jpg')),
        })

        doc.save(self.generated_path)

        self.assertEqual(len(doc.get_field_names()), 0)
        self.assertEqual(len(doc._element.xpath('w:body/w:p/w:r/w:drawing')), 1)

        self.assertTrue(any(re.match('word/media/.*\.jpg', f.filename)
                            for f in ZipFile(self.generated_path).infolist()))


class TableReplacement(DocxTestsBase):
    def test_replace_table(self):
        doc = self._get_docx('table')

        doc.replace_fields({
            'table': Table([[str(i * j) for j in range(1, 5)] for i in range(1, 3)],
                           headers=['one', 'two', 'three', 'four']),
        })

        self.assertEqual(len(doc.get_field_names()), 0)

        self.assertTrue(doc._element.xpath('.//w:tbl'))
        self.assertEqual(len(doc._element.xpath('.//w:tbl/w:tr')), 3)
        self.assertEqual(len(doc._element.xpath('(.//w:tbl//w:tr)[1]/w:tc')), 4)
        self.assertEqual(doc._element.xpath('((.//w:tbl//w:tr)[1]/w:tc//w:t//text())'), ['one', 'two', 'three', 'four'])
        self.assertEqual(doc._element.xpath('((.//w:tbl//w:tr)[3]/w:tc//w:t//text())'), ['2', '4', '6', '8'])

    def test_replace_table_no_headers(self):
        doc = self._get_docx('table')

        doc.replace_fields({
            'table': Table([[str(i * j) for j in range(1, 5)] for i in range(1, 3)], headers=None)
        })

        self.assertEqual(len(doc.get_field_names()), 0)

        self.assertTrue(doc._element.xpath('.//w:tbl'))
        self.assertEqual(len(doc._element.xpath('.//w:tbl/w:tr')), 2)
        self.assertEqual(len(doc._element.xpath('(.//w:tbl//w:tr)[1]/w:tc')), 4)
        self.assertEqual(doc._element.xpath('((.//w:tbl//w:tr)[2]/w:tc//w:t//text())'), ['2', '4', '6', '8'])

    def test_malformed_table(self):
        doc = self._get_docx('table')

        rows = [[str(i * j) for j in range(1, 5)] for i in range(1, 3)]
        rows.append(['this row is not the same length'])

        with self.assertRaises(ValueError) as cm:
            doc.replace_fields({
                'table': Table(rows,
                               headers=['one', 'two', 'three', 'four']),
            })

            self.assertEqual(cm.exception.message, 'Data contains rows of varying sizes')

    def test_malformed_headers(self):
        doc = self._get_docx('table')

        rows = [[str(i * j) for j in range(1, 5)] for i in range(1, 3)]
        rows.append(['this row is not the same length'])

        with self.assertRaises(ValueError) as cm:
            doc.replace_fields({
                'table': Table(rows,
                               headers=['one', 'two', 'three']),
            })

            self.assertEqual(cm.exception.message, 'Data and header lengths do not match')


class HTMLReplacementTests(DocxTestsBase):
    def test_html_replacement(self):
        doc = self._get_docx('html')

        html = """
        <h1>Header</h1>

        <h2>Subheader</h2>

        <p><strong>bold</strong> - not bold</p>

        <ul>
            <li>hop</li>
            <li>la</li>
            <li>kee</li>
        </ul>"""

        doc.replace_fields({
            'html': HTML(html)
        })

        self.assertEqual(len(doc.get_field_names()), 0)

        self.assertTrue(doc._element.xpath('.//text()="Subheader"'))
        self.assertFalse(doc._element.xpath('.//text()="h1"'))

        # the run with the "bold" text in it has an rPr tag with a b tag in it
        self.assertTrue(doc._element.xpath('.//*[text()="bold"]/parent::w:r/w:rPr/w:b'))

        # the run with the "not bold" text in it exists but does not have a rPr with a b tag in it
        self.assertTrue(doc._element.xpath('.//*[contains(text(), "not bold")]/parent::*'))
        self.assertFalse(doc._element.xpath('.//*[contains(text(), "not bold")]/parent::*/w:b'))

        # assert that all styles mentioned in the document are also in the document's style section
        style_ids = doc.element.xpath('.//w:pStyle/@w:val')
        self.assertTrue(all(doc.styles.get_style_id(sid, WD_STYLE_TYPE.PARAGRAPH) for sid in style_ids))



from bureaucracy.replacements import HTMLReplacement, TextReplacement, ImageReplacement, TableReplacement
from bureaucracy.template import DocxTemplate

HTML = HTMLReplacement
Text = TextReplacement
Image = ImageReplacement
Table = TableReplacement

__all__ = ['DocxTemplate',
           'HTML',
           'Text',
           'Table',
           'Image',
           'Document']
